/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

/** \file soc.h
    \brief Stochastic Optimal Control library */

#ifndef MT_soc_h
#define MT_soc_h

#include "array.h"

//-- fwd declarations
class OpenGL;
struct SwiftModule;
namespace ors{ struct Graph; }
struct ControlVariable;
typedef MT::Array<ControlVariable*> ControlVariableList;


namespace soc{

/// gradient method options
enum { ConjGrad=0, LevMar=1, Rprop=2, RpropConjGrad=3, SQP=4, Attractor=5 };


//===========================================================================
//
// SocAbstraction
//

/** \brief defines an abstraction of stochastic optimal control
    problems -- see section 3.2 of the <a
    href="../guide.pdf">guide</a> */
struct SocAbstraction{

  ///@name data fields
  std::ostream *os; ///< if non-NULL, some routines might give output
  OpenGL *gl;       ///< if non-NULL, some routines might give output
  bool dynamic;     ///< determines whether this problem is dynamic or not

  ///@name initialization
  SocAbstraction();
  virtual ~SocAbstraction();

  ///@name low level access routines: need to be implemented by the simulator

  // general problem information
  virtual uint nTime() = 0;          ///< total time of the trajectory
  virtual uint nTasks() = 0;         ///< number of task variables
  virtual uint qDim() = 0;           ///< dimensionality of q-space
  virtual uint uDim() = 0;           ///< dimensionality of control
  virtual uint yDim(uint i) = 0;     ///< dimensionality of the i-th task
  virtual void getq0 (arr& q) = 0;   ///< start joint configuration
  virtual void getqv0(arr& q_) = 0;  ///< start joint configuration and velocity
  virtual void getqv0(arr& q,arr& qd) = 0;  ///< start joint configuration and velocity
  virtual double getTau() = 0;       ///< time step size (for dynamic problems)
  void getx0(arr& x){ if(dynamic) getqv0(x); else getq0(x); }

  // set x-state (following calls to getPhi and getJ are w.r.t. this x)
  virtual void setq  (const arr& q) = 0;
  virtual void setqv (const arr& q_) = 0;
  virtual void setqv (const arr& q,const arr& qd) = 0;
  void setx(const arr& x){ if(dynamic) setqv(x); else setq(x); }

  //motion prior, or control cost
  //virtual void geth  (arr& h) = 0; NIY...
  virtual void getW  (arr& W) = 0;
  virtual void getH  (arr& H) = 0;
  virtual void getQ  (arr& Q) = 0;

  // task coupling
  virtual bool isConditioned(uint i,uint t) = 0;
  virtual void getPhi(arr& phiq_i,uint i) = 0;
  virtual void getJtJ(arr& J_i,arr& tJ_i,uint i) = 0;
  virtual void getJqd(arr& Jqd_i,uint i) = 0;
  virtual void getTarget (arr& y_i,uint i,uint t) = 0;
  virtual void getTargetV(arr& v_i,uint i,uint t) = 0;
  //virtual void getC   (arr& C_i,uint i,uint t) = 0;
  //virtual void getCV  (arr& D_i,uint i,uint t) = 0;
  virtual void getPrecision (double& prec,uint i,uint t) = 0;
  virtual void getPrecisionV(double& prec,uint i,uint t) = 0;
  virtual void getHessian(arr& H_i,uint i); //{ NIY; }

  // dynamic model
  virtual void   getMF(arr& M,arr& F) = 0;
  virtual void   getMinvF(arr& Minv,arr& F) = 0;


  ///@name higher level methods, access only the upper methods

  // abstract SOC interface
  void getProcess(arr& A,arr& a,arr& B);
  void getProcess(arr& A,arr& tA,arr& invA,arr& invtA,arr& a,arr& B,arr& tB);
  double getCosts(arr& R,arr& r,uint t,const arr& qt);

  // cost info
  double computeTotalCost(const arr& q,bool plot=true);
  void computeTotalCostGradient(arr& dCdq,const arr& q);
  
  //-- convenience (prelim...)
  void displayUpdate(const char *text=NULL,bool wait=false);
  void displayTrajectory(const arr& q,int display,const char *tag=NULL);
  double analyzeDynamicTrajectory(const arr& q);
  void constantTrajectory(arr& q);
  void passiveDynamicTrajectory(arr& q);
  void controlledDynamicTrajectory(arr& q,const arr& u);
  void getControlFromTrajectory(arr& u,const arr& q);
};


//===========================================================================
//
///@name     trivial helpers
// @{

void getVelocity(arr& vt,const arr& q,uint t,double tau);
void getPhaseTrajectory(arr& _q,const arr& q,double tau);
void getPositionTrajectory(arr& q,const arr& _q);

//only for the first task so far!
void getJointFromTaskTrajectory(SocAbstraction& soci,arr& q,const arr& x);
void partialJointFromTaskTrajectory(SocAbstraction& soci,arr& dx,const arr& dq,const arr& q,const arr& x);
void straightTaskTrajectory(SocAbstraction& soci,arr& q,uint taskid);


//===========================================================================
// @}
///@name     inverse kinematics control
// @{

void bayesianIKControl(SocAbstraction& soci, arr& dq, uint t);
void pseudoIKControl(SocAbstraction& soci, arr& dq, uint t,double regularization=1e-8);
void hierarchicalIKControl(SocAbstraction& soci, arr& dq, uint t,double regularization=1e-8);
void bayesianIterateIKControl(SocAbstraction& soci,
                              arr& qt,const arr& qt_1,uint t,double eps,uint maxIter);
void bayesianIKTrajectory(SocAbstraction& soci, arr& q, double eps=-1);
void bayesianDynamicControl(SocAbstraction& soci, arr& qv, const arr& qv_1, uint t);



//===========================================================================
// @}
///@name     gradient optimization
// @{

void gradientOptimization(SocAbstraction& soci,
                                arr& q,
                                int gradient_method,
                                uint maxIterations,
                                uint spline_points,
                                uint spline_degree,
                                double stoppingTolerance,
                                bool checkGradient,
                                uint display);

void gradientTaskOptimization(SocAbstraction& soci,
                                arr& q,
                                uint spline_points,
                                uint spline_degree,
                                int gradient_method,
                                uint maxIterations,
                                double stoppingTolerance,
                                bool checkGradient,
                                uint display);

void gradientAttractorTaskOptimization(SocAbstraction& soci,
                                arr& q,
                                uint spline_points,
                                uint spline_degree,
                                int gradient_method,
                                uint maxIterations,
                                double stoppingTolerance,
                                bool checkGradient,
                                uint display);

void SQPOptimization(SocAbstraction& soci,
                           arr& q,uint iterations,
                           uint spline_points,
                           uint spline_degree,
                           uint display);

//===========================================================================
// @}
///@name     iLQG methods
// @{

void iLQG_kinematic(SocAbstraction& soci,
                           arr& q,uint iterations,
                           double convergenceRate,
                           uint display);

void iLQG_general(SocAbstraction& soci,
                  arr& q,uint iterations,
                  double convergenceRate,
                  uint display);
 

//===========================================================================
// @}
///@name     bayesian inference methods
// @{

void AICO_kinematic(SocAbstraction& soci,
		    arr& q,uint iterations,
		    double stepSmooth,double repeatThreshold,
		    uint display);

void AICO_general(SocAbstraction& soci,
		  arr& q,uint iterations,
		  double convergenceRate,double repeatThreshold,
		  uint display);

void BayesianDynamicMotionPlanning_obsolete(SocAbstraction& soci,
                                     arr& q,uint iterations,
                                     double stepSmooth,double repeatThreshold,
                                     uint display);

void BayesianDynamicFilter_obsolete(SocAbstraction& soci,
				    arr& q,uint iterations,
				    double stepSmooth,double repeatThreshold,
				    uint display);

//===========================================================================
// @}
///@name preliminary or obsolete
// @{

class OrsSocImplementation;
class ToySocImplementation;

double getDynamicCostMeassure(SocAbstraction& soci,arr& q,double& cost1,double& cost2,std::ostream *os=0);

double getFilterCostMeassure(SocAbstraction& soci,arr& q,double& cost1,double& cost2,std::ostream *os=0);

double getTaskCost(soc::SocAbstraction& soci,int t);
void getTaskCostGradient(soc::SocAbstraction& soci,arr& dCdq,int t);
void getQuadraticTaskCost(soc::SocAbstraction& soci,int t,arr& R,arr& r,const arr& qt);

void setupOpenGL(OrsSocImplementation &soci);

void createEndeffectorReachProblem(OrsSocImplementation &soci,
                                   const char *ors_file,
                                   uint trajectory_length,
                                   int rand_seed);

/*void createNikolayReachProblem(OrsSocImplementation &soci,
                               slGraph &ors,
                               SwiftModule& swift,
                               uint trajectory_length,
                               const arr& endeffector_target,
                               const char* endeffector_name,
                               const arr& W);*/

void createDynamicProblem(OrsSocImplementation &soci,
                          const char *ors_file,
                          double trajectory_time,
                          uint trajectory_steps);

void setupOpenGL(ToySocImplementation &soci);

void createEndeffectorReachProblem(ToySocImplementation &soci,
                                   const char *ors_file,
                                   uint trajectory_length,
                                   int rand_seed);

/*void createNikolayReachProblem(ToySocImplementation &soci,
                               slGraph &ors,
                               SwiftModule& swift,
                               uint trajectory_length,
                               const arr& endeffector_target,
                               const char* endeffector_name,
                               const arr& W);*/

void createDynamicProblem(ToySocImplementation &soci,
                          const char *ors_file,
                          double trajectory_time,
                          uint trajectory_steps);

// @}

//===========================================================================
//
// ORS simulator implementation of the SocAbstration
//

struct OrsSocWorkspace;

/** \brief an implementation of the SocAbstraction using the \ref ors
    simulator */
struct OrsSocImplementation: public virtual SocAbstraction{
  OrsSocWorkspace *WS;

  OrsSocImplementation();
  virtual ~OrsSocImplementation();

  //initialization methods
  void initKinematic(ors::Graph *ors,SwiftModule *swift,OpenGL *gl,uint trajectory_length);
  void initPseudoDynamic(ors::Graph *ors,SwiftModule *swift,OpenGL *gl,
                         double trajectory_time,uint trajectory_steps);
  void setControlVariables(const ControlVariableList& CVlist);

  //info
  void reportOnState(std::ostream& os);

  //implementations of virtual methods
  uint nTime();
  uint nTasks();
  uint qDim();
  uint uDim();
  uint yDim(uint i);
  void getq0 (arr& q);
  void getqv0(arr& q_);
  void getqv0(arr& q,arr& qd);
  bool isDynamic();
  void setq  (const arr& q);
  void setqv (const arr& q_);
  void setqv (const arr& q,const arr& qd);
  void geth  (arr& h);
  void getW  (arr& W);
  void getH  (arr& H);
  void getQ  (arr& Q);
  bool isConditioned(uint i,uint t);
  void getPhi(arr& phiq_i,uint i);
  void getJtJ(arr& J_i,arr& tJ_i,uint i);
  void getHessian(arr& H_i,uint i);
  void getJqd(arr& Jqd_i,uint i);
  void getTarget (arr& y_i,uint i,uint t);
  void getTargetV(arr& v_i,uint i,uint t);
  void getC   (arr& C_i,uint i,uint t);
  void getCV  (arr& D_i,uint i,uint t);
  void getPrecision (double& prec,uint i,uint t);
  void getPrecisionV(double& prec,uint i,uint t);
  double getTau();
  void getMF(arr& M,arr& F);
  void getMinvF(arr& Minv,arr& F);
};

//========= untidy stuff ...



//===========================================================================
//
// toy implementation of the SocAbstration
//

struct ToySocWorkspace;

/** \brief an implementation of the SocAbstraction that simulates a
    single 1D point mass on a spring */
struct ToySocImplementation: public virtual SocAbstraction{
  ToySocWorkspace *WS;

  ToySocImplementation();
  virtual ~ToySocImplementation();

  //implementations of virtual methods
  uint nTime();
  uint nTasks();
  uint qDim();
  uint uDim();
  uint yDim(uint i);
  void getq0 (arr& q);
  void getqv0(arr& q_);
  void getqv0(arr& q,arr& qd);
  bool isDynamic();
  void setq  (const arr& q);
  void setqv (const arr& q_);
  void setqv (const arr& q,const arr& qd);
  void geth  (arr& h);
  void getW  (arr& W);
  void getH  (arr& H);
  void getQ  (arr& Q);
  bool isConditioned(uint i,uint t);
  void getPhi(arr& phiq_i,uint i);
  void getJtJ(arr& J_i,arr& tJ_i,uint i);
  void getJqd(arr& Jqd_i,uint i);
  void getTarget (arr& y_i,uint i,uint t);
  void getTargetV(arr& v_i,uint i,uint t);
  void getC   (arr& C_i,uint i,uint t);
  void getCV  (arr& D_i,uint i,uint t);
  void getPrecision (double& prec,uint i,uint t);
  void getPrecisionV(double& prec,uint i,uint t);
  double getTau();
  void getMF(arr& M,arr& F);
  void getMinvF(arr& Minv,arr& F);
};

}


//===========================================================================
//
// implementations
//

#ifdef MT_IMPLEMENTATION
#  include "soc.cpp"
#  include "soc_methods.cpp"
#  include "soc_methods_gradient.cpp"
#  include "soc_methods_bayes.cpp"
#  include "soc_methods_lqg.cpp"
#  include "soc_orsImplementation.cpp"
#  include "soc_toyImplementation.cpp"
//#  include "soc_methods_attractor.cpp"
#endif

#endif
