/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "soc.h"
#include "util.h"

//===========================================================================
//
// helpers - fwd declarations
//

void setBlockVec(arr& X, double a0,double a1, const arr& x);
void setBlockVec(arr& X,const arr& x1,const arr& x2);
void setBlockDiagMat(arr& X,const arr& A,const arr& B);
void setBlockMat(arr& X, double a00,double a01,double a10,double a11, const arr& x);
void blockTimesVec(arr& Y, double a00,double a01,double a10,double a11, const arr& X);
void bl_ockTimesMat(arr& Y, double a00,double a01,double a10,double a11, const arr& X);
void blockTransMat(arr& Y, double a,double b,double c,double d, const arr& X);
double getDynamicCostMeassure(soc::SocAbstraction& soci,arr& q,double& cost1,double& cost2,std::ostream *os);
double getTransitionLogLikelihood(soc::SocAbstraction& soci,arr& b0,arr& b1);

//===========================================================================
//
// methods
//

/*! \brief compute a single control step from current state to state
    at time t (taking into account all targets and precisions
    associated with all declared task variables) */
void soc::bayesianIKControl(SocAbstraction& soci,
                       arr& dq,uint t){
  uint n=soci.qDim(),m=soci.nTasks();
  dq.resize(n);
  dq.setZero();
  arr a(n),A(n,n),invA(n,n);
  arr target,actual,J,tJ;
  double prec;
  soci.getW(A);
  a.setZero();
  //arr w(3);
  for(uint i=0;i<m;i++) if(soci.isConditioned(i,t)){
    soci.getPhi   (actual,i);
    soci.getJtJ   (J,tJ,i);
    soci.getTarget(target,i,t);
    soci.getPrecision(prec,i,t);
    a += prec * tJ * (target-actual);
    A += prec * tJ * J;
  }
  inverse_SymPosDef(invA,A);
  dq = invA * a;
}

/*! \brief standard IK -- follows the first (active) task variable
    exactly, but fails if more than one task variable is
    active. regularization=make it singularity robust*/
void soc::pseudoIKControl(SocAbstraction& soci, arr& dq, uint t,double regularization){
  uint n=soci.qDim(),m=soci.nTasks();
  dq.resize(n);
  dq.setZero();
  arr invJ,W,invW;
  arr target,actual,J,tJ;
  soci.getW(W);
  inverse_SymPosDef(invW,W);
  uint k=0;
  for(uint i=0;i<m;i++) if(soci.isConditioned(i,t)){
    CHECK(!k,"pseudo IK only works for a single task variable");
    soci.getPhi   (actual,i);
    soci.getJtJ   (J,tJ,i);
    soci.getTarget(target,i,t);
    pseudoInverse(invJ,J,invW,regularization);
    dq = invJ * (target-actual);
    k++;
  }
}

/*! \brief hierarchical IK: follows the 1st task variable exactly, the
    2nd exactly up to the 1st, etc.., might be come
    brittle. regularization=make it singularity robust */
void soc::hierarchicalIKControl(SocAbstraction& soci,arr& dq,uint t,double regularization){
  uint n=soci.qDim(),m=soci.nTasks();
  dq.resize(n);
  dq.setZero();
  arr hatJ,invHatJ,N,W,invW;
  arr target,actual,J,tJ;
  N.setId(n);
  soci.getW(W);
  inverse_SymPosDef(invW,W);
  for(uint i=0;i<m;i++) if(soci.isConditioned(i,t)){
    soci.getPhi   (actual,i);
    soci.getJtJ   (J,tJ,i);
    soci.getTarget(target,i,t);
    hatJ = J * N;
    pseudoInverse(invHatJ,hatJ,invW,regularization);
    dq += invHatJ * ((target-actual) - J * dq);
    N  -= invHatJ * hatJ;
  }
}

//===========================================================================

/*! \brief compute a single control step from state at time t-1 to
    state at time t. If eps=0, this is equivalen to bayesianIKControl;
    for eps>0 the IK step is repeated until convergence up to
    tolerance eps. qt=output, qt_1=state at time t-1. */
void soc::bayesianIterateIKControl(SocAbstraction& soci,
                                   arr& qt,const arr& qt_1,uint t,double eps,uint maxIter){
  uint j;
  if(&qt!=&qt_1) qt=qt_1;
  arr dq;
  for(j=0;j<maxIter;j++){
    soci.setq(qt);
    bayesianIKControl(soci,dq,t);
    if(j<3) qt+=dq;
    //else if(j<20) qt+=.8*dq;
    else qt+=.8*dq;
    if(dq.absMax()<eps) break;
    //cout <<"IK iteration " <<j <<" dq=" <<dq <<endl;
  }
  if(j==maxIter) HALT("warning: IK didn't converge (|last step|="<<dq.absMax()<<")")
  else cout <<"IK converged after steps=" <<j <<endl;
}

//===========================================================================


/*! \brief compute a single control step from current state to target of time t.
    qv=output, qv_1=state at time t-1 */
void soc::bayesianDynamicControl(SocAbstraction& soci, arr& qv, const arr& qv_1, uint t){
  CHECK(soci.dynamic,"assumed dynamic SOC abstraction");
  uint n=soci.qDim();
  
  //-- access necessary information
  arr A,a,B,tB;
  soci.getProcess(A,a,B);
  transpose(tB,B);
  
  arr Q,H,invH;
  soci.getQ(Q);
  soci.getH(H);
  inverse_SymPosDef(invH,H);
  
  //fwd message
  arr s(2*n),S(2*n,2*n),invS(2*n,2*n);
  S = Q;
  S += B*invH*tB;
  s = a + A*qv_1;
  inverse_SymPosDef(invS,S);

  //task message
  arr R,r,q_1;
  q_1.referToSubRange(qv_1,0,n-1);
  soci.getCosts(R,r,0,q_1);
  r *= -.5;

  //belief
  arr invB,b;
  invB = invS + R;
  lapack_Ainv_b_sym(b, invB, invS*s + r);

  qv=b;
}
                           

//===========================================================================

/*! \brief compute a trajectory using inverse kinematics (iterating
    bayesianIK forward). If eps=0, no IK step is repeated, for eps>0
    IK steps are repeated until they converge up to tolerance eps (see
    \ref bayesianIterateIKControl) */
void soc::bayesianIKTrajectory(SocAbstraction& soci,arr& q,double eps){
  uint t,T=soci.nTime(),n=soci.qDim();
  q.resize(T,n);
  arr dq;
  soci.getq0(q[0]());
  for(t=1;t<T;t++){
    if(eps<0){
      soci.setq(q[t-1]);
      //soci.gl->watch();
      bayesianIKControl(soci,dq,t);
      q[t]() = q[t-1] + dq;
    }else{
      bayesianIterateIKControl(soci,q[t](),q[t-1],t,eps,20);
      //soci.gl->watch();
    }
    //cout <<q[t] <<endl;
    //reportAll(globalSpace,cout);
  }
}

/*
void bayesianIterateIKTrajectory(SocAbstraction& soci,arr& q,double eps,uint maxIter){
  uint t,T=soci.nTime(),n=soci.qDim();
  q.resize(T,n);
  arr dq;
  soci.getq0(q[0]());
  for(t=1;t<T;t++){
    bayesianIterateIKControl(soci,q[t](),q[t-1],t,eps,maxIter);
}
}
*/


//===========================================================================

//#define CANON

void lapack_A_Binv_A_sym(arr& X,const arr& A,const arr& B){
  static arr Binv,tmp;
  inverse_SymPosDef(Binv,B);
  blas_MM(tmp,Binv,A);
  blas_MM(X,A,tmp);
}

//! Approximate Inference Control (AICO) in the kinematic case
void soc::AICO_kinematic(SocAbstraction& soci,
			 arr& q,uint iterations,
			 double convergenceRate,double repeatThreshold,
			 uint display){
  CHECK(!soci.dynamic,"assumed dynamic SOC abstraction");
  uint n=soci.qDim();
  uint T=soci.nTime();
  uint k,t,t0;
  int dt;

  if(iterations%2) MT_MSG("you should always choose even number of fwd/bwd iterations!");
  
  bool useInitialTrajectory=false;
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"initial trajectory was wrong dimensionality");
    useInitialTrajectory=true;
    if(soci.os){//type initial value
      *soci.os <<std::setw(3) <<-1 <<" time " <<MT::timerRead(false);
      soci.computeTotalCost(q);
    }
  }

  //general variables
  arr q0,W,invW,Id;
  soci.getW(W);
  inverse_SymPosDef(invW,W);
  Id.setId(n);
  soci.getq0(q0);

  //variables to store beliefs
  arr barS,barV,//tmp2,tmp3,
    s(T,n),invS(T,n,n),
    v(T,n),invV(T,n,n),
    b(T,n),invB(T,n,n),
    r(T,n),R(T,n,n),
    hatq(T,n);
  s[0]=q0;      invS[0].setDiag(1e10);
#ifdef CANON
  s[0]=invS[0]*q0;
#endif
  v.setZero();  invV.setZero();
  b[0]=q0;      invB[0].setDiag(1e10);
  r[0]=0.;      R[0]=0.;

  MT::timerStart();
  uint repeatCount=0;

  for(k=0;k<iterations;k++){
    if(!(k&1)){ dt=1; t0=1; }else{ dt=-1; t0=T-1; }
    for(t=t0;t<T && t>0;t+=dt){
#ifndef CANON //Gaussians in natural form
      //compute (s,S)
      inverse_SymPosDef(barS,invS[t-1] + R[t-1]);
      s[t] = barS * (invS[t-1]*s[t-1] + r[t-1]);
      inverse_SymPosDef(invS[t](), invW + barS);
      //cout <<"s\n" <<s[t] <<endl <<invS[t] <<endl;

      //compute (v,V)
      if(k && t<T-1){
        inverse_SymPosDef(barV,invV[t+1] + R[t+1]);
        v[t] = barV * (invV[t+1]*v[t+1] + r[t+1]);
        inverse_SymPosDef(invV[t](), invW + barV);
      }
      if(k && t==T-1){
        v[t] = b[t];
        //invV[t].setDiag(1e10); //fixes the end posture!, use 1e-5 otherwise [[HACK]]
        invV[t].setDiag(1e0);
      }
      if(!k && useInitialTrajectory){
        v[t]() = q[t];
        invV[t].setDiag(1e0);
      }
      //cout <<"v\n" <<v[t] <<endl <<invV[t] <<endl;
      
      //decide on \hat q
      if(k>0 && convergenceRate){
        hatq[t]()=(1.-convergenceRate)*hatq[t] + convergenceRate*b[t];
      }else{
        if(k)
          hatq[t]()=b[t];
        else{
          if(useInitialTrajectory){
            hatq[t]()=q[t];
          }else{
            hatq[t]()=s[t];
          }
        }
      }
      if(k==1) hatq[t]()=v[t];
      
      //compute (r,R)
      soci.setq(hatq[t]);
      //soci.displayUpdate(STRING("kinematic AICO, iteration "<<k <<", time step "<<t),true);
      //getQuadraticTaskCost(soci,t,R[t](),r[t](),hatq[t]);
      soci.getCosts(R[t](),r[t](),t,hatq[t]);
      r[t]() *= -.5;
      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;

      //compute (b,B);
      invB[t] = invS[t] + invV[t] + R[t];
      lapack_Ainv_b_sym(b[t](), invB[t], invS[t]*s[t] + invV[t]*v[t] + r[t]);
      //cout <<"b\n" <<b[t] <<endl <<B[t] <<endl;
#else //Gaussians in canonical form
      //compute (s,S)
      //if(dt==1){
      barS = invS[t-1] + R[t-1];
      //lapack_A_Binv_A_sym(invS[t](),barS,-(barS+W));
      //invS[t]() += barS;
      inverse_SymPosDef(tmp2,barS+W);
      blas_MM(tmp3,barS,tmp2);
      tmp2 = Id - tmp3;//*tmp2;
      blas_MM(invS[t](),tmp2,barS);
      //lapack_invA_b_sym
      s[t] = tmp2*(s[t-1] + r[t-1]);
      //cout <<"s\n" <<s[t] <<endl <<invS[t] <<endl;
      //}

      //compute (v,V)
      //if(dt==-1){
      if(k && t<T-1){
        inverse_SymPosDef(barV,invV[t+1] + R[t+1]);
        v[t] = barV * (invV[t+1]*v[t+1] + r[t+1]);
        inverse_SymPosDef(invV[t](), invW + barV);
      }
      if(k && t==T-1){
        v[t] = b[t];
        //invV[t].setDiag(1e10); //fixes the end posture!, use 1e-5 otherwise [[HACK]]
        invV[t].setDiag(1e0);
      }
      if(!k && useInitialTrajectory){
        v[t]() = q[t];
        invV[t].setDiag(1e0);
      }
      //}
      //cout <<"v\n" <<v[t] <<endl <<invV[t] <<endl;
      
      //decide on \hat q
      if(k>0 && convergenceRate){
        hatq[t]()=(1.-convergenceRate)*hatq[t] + convergenceRate*b[t];
      }else{
        if(k)
          hatq[t]()=b[t];
        else{
          if(useInitialTrajectory){
            hatq[t]()=q[t];
          }else{
            lapack_Ainv_b_sym(hatq[t](),invS[t],s[t]);
          }
        }
      }

      //compute (r,R)
      soci.setq(hatq[t]);
      getQuadraticTaskCost(soci,t,R[t](),r[t](),hatq[t]);
      r[t]() *= -.5;
      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;

      //compute (b,B);
      invB[t] = invS[t] + invV[t] + R[t];
      lapack_Sinv_b_sym(b[t](), invB[t], s[t] + invV[t]*v[t] + r[t]);
      //cout <<"b\n" <<b[t] <<endl <<B[t] <<endl;
#endif

      if(k>1 && repeatThreshold){
        double off=sqrDistance(W,b[t],hatq[t]);
        if(off>repeatThreshold){
          //cout <<t <<" REPEAT: off=" <<off <<" (repeatCount=" <<repeatCount <<")" <<endl;
          if(repeatCount<20){
            t-=dt;
            repeatCount++;
          }else{
            cout <<" ** no convergence! ** (skipping repeat)" <<endl;
            repeatCount=0;
          }
        }else{
          repeatCount=0;
        }
      }
    }

    //display or evaluate
    if(soci.os){
      *soci.os <<std::setw(3) <<k <<" time " <<MT::timerRead(false);
      soci.computeTotalCost(b);
    }
    if(soci.gl){
      soci.displayTrajectory(b,display,STRING("AICO_kinematic - iteration "<<k));
    }
  }

  q = b;
}

//! Approximate Inference Control (AICO) in the general (e.g. dynamic) case
void soc::AICO_general(SocAbstraction& soci,
		       arr& q,uint iterations,
		       double convergenceRate,double repeatThreshold,
		       uint display){
  CHECK(soci.dynamic,"assumed dynamic SOC abstraction");
  uint n=soci.qDim();
  uint T=soci.nTime();
  uint n2=2*n;
  uint k,t,t0;
  int dt;

  if(iterations%2) MT_MSG("you should always choose even number of fwd/bwd iterations!");

  bool useInitialTrajectory=false;
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"initial trajectory was wrong dimensionality");
    useInitialTrajectory=true;
    if(soci.os){//type initial value
      *soci.os <<std::setw(3) <<-1 <<" time " <<MT::timerRead(false);
      soci.computeTotalCost(q);
    }
    if(soci.gl){
      soci.displayTrajectory(q,display,STRING("AICO_general - iteration -1"));
    }
  }

  //variables for the dynamics
  arr q0;
  arr A(T,n2,n2),tA(T,n2,n2),invA(T,n2,n2),invtA(T,n2,n2),a(T,n2),B(T,n2,n),tB(T,n,n2); //fwd dynamics
  arr Q,H,invH;
  soci.getqv0(q0);
  soci.getQ(Q);
  soci.getH(H);
  inverse_SymPosDef(invH,H);
  CHECK(q0.N==n2,"");
  soci.setqv(q0);
  soci.getProcess(A[0](),tA[0](),invA[0](),invtA[0](),a[0](),B[0](),tB[0]());

  //variables to store beliefs
  arr St,Vt,barS,barV,invR,
    s(T,n2),invS(T,n2,n2),
    v(T,n2),invV(T,n2,n2),
    b(T,n2),invB(T,n2,n2),
    r(T,n2),R(T,n2,n2),
    hatq(T,n2);
  s[0]=q0;      invS[0].setDiag(1e10);
  v.setZero();  invV.setZero();
  b[0]=q0;      invB[0].setDiag(1e10);
  r[0]=0.;      R[0]=0.;

  if(useInitialTrajectory) getPhaseTrajectory(b,q,soci.getTau());

  MT::timerStart();
  uint repeatCount=0;
  
  for(k=0;k<iterations;k++){
    if(!(k&1)){ dt=1; t0=1; }else{ dt=(uint)-1; t0=T-1; }
    for(t=t0;t<T && t>0;t+=dt){
      //compute (s,S)
      inverse_SymPosDef(barS,invS[t-1] + R[t-1]);
      St = Q;
      St += B[t-1]*invH*tB[t-1];
      St += A[t-1]*barS*tA[t-1];
      s[t] = a[t-1] + A[t-1]*(barS*(invS[t-1]*s[t-1] + r[t-1]));
      inverse_SymPosDef(invS[t](),St);
      //cout <<"s\n" <<s[t] <<endl <<invS[t] <<endl;

      //decide on \hat q
      if(k>0 && convergenceRate){
        hatq[t]()=(1.-convergenceRate)*hatq[t] + convergenceRate*b[t];
      }else{
        if(k)
          hatq[t]()=b[t];
        else{
          if(useInitialTrajectory){
	    NIY;
            hatq[t]()=b[t]; //perhaps encode initial trajectory only in backward messages
          }else{
            hatq[t]()=s[t];
          }
        }
      }
      //if(k==1) hatq[t]()=v[t];

      //compute (r,R)
      soci.setqv(hatq[t]);
      soci.getCosts  (R[t](),r[t](),t,hatq[t].sub(0,n-1));
      soci.getProcess(A[t](),tA[t](),invA[t](),invtA[t](),a[t](),B[t](),tB[t]());
      r[t]() *= -.5;
      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;

      //compute (v,V)
      if(k && t<T-1){
        inverse_SymPosDef(barV,invV[t+1] + R[t+1]);
        Vt = Q;
        Vt += B[t]*invH*tB[t];
        Vt += barV;
        Vt = invA[t]*Vt*invtA[t];
        //cout <<barV <<invH <<Vt <<invA[t] <<invtA[t] <<B[t] <<tB[t]  <<endl;
        v[t] = -invA[t]*a[t] + invA[t]*barV*(invV[t+1]*v[t+1] + r[t+1]);
        inverse_SymPosDef(invV[t](), Vt);
      }else{
        if(!k && useInitialTrajectory){ //first fwd pass
          v[t]() = b[t];
          invV[t].setDiag(1e1);
        }
        if(k && t==T-1){ //last time slice
          v[t] = b[t];
          invV[t].setDiag(1e0);
        }
      }
      //cout <<"v\n" <<v[t] <<endl <<invV[t] <<endl;

      arr invD,d;
      double like;
      do{
      //compute (b,B);
      invB[t] = invS[t] + invV[t] + R[t];
      lapack_Ainv_b_sym(b[t](), invB[t], invS[t]*s[t] + invV[t]*v[t] + r[t]);
      //cout <<"b\n" <<b[t] <<endl <<invB[t] <<endl;

      //compute likelihood
      invD = invS[t] + invV[t];
      lapack_Ainv_b_sym(d, invD, invS[t]*s[t] + invV[t]*v[t]);
      like=metricDistance(invD,d,b[t]);
      /*cout <<t
        <<" - fwd like=" <<metricDistance(invS[t],s[t],b[t])
        <<" - bwd like=" <<metricDistance(invV[t],v[t],b[t])
        <<" - r like=" <<like
        <<endl;
      */
      if(like>100.){
        cout <<" REDUCING task precision" <<endl;
        R[t]() *= .1;
        r[t]() *= .1;
      }
      }while(like>100.);

      if(k>0 && repeatThreshold){
        //double off=sqrDistance(Q,b[t],hatq[t]);
        double off=sqrDistance(b[t],hatq[t]);
        if(off>repeatThreshold){
          //cout <<t <<" REPEAT: off=" <<off <<" (repeatCount=" <<repeatCount <<")" <<endl;
          if(repeatCount<20){
            t-=dt;
            repeatCount++;
          }else{
            cout <<" ** no convergence! ** (skipping repeat)" <<endl;
            repeatCount=0;
          }
        }else{
          repeatCount=0;
        }
      }
    }

    //display or evaluate
    if(soci.os){
      *soci.os <<std::setw(3) <<k <<" time " <<MT::timerRead(false);
      getPositionTrajectory(q,b);
      //soci.analyzeDynamicTrajectory(q);
      soci.computeTotalCost(q);
    }
    if(soci.gl){
      if(!soci.os) getPositionTrajectory(q,b);
      //do{
      soci.displayTrajectory(q,display,STRING("AICO_general - iteration "<<k));
      //}while(soci.gl->watch()!=32);
    }
  }

  getPositionTrajectory(q,b);
}

//===========================================================================

//! obsolete
void soc::BayesianDynamicMotionPlanning_obsolete(SocAbstraction& soci,
						 arr& q,uint iterations,
						 double stepSmooth,double repeatThreshold,
						 uint display){
  uint n=soci.qDim();
  uint T=soci.nTime();
  uint m=soci.nTasks();
  double tau=soci.getTau(), tau2=tau*tau;

  //general variables
  arr q_0,W,invW,H,invH,Q;
  soci.getW(W);
  soci.getH(H);
  soci.getQ(Q);
  inverse_SymPosDef(invW,W);
  inverse_SymPosDef(invH,H);
  soci.getqv0(q_0);

  //variables that describe the constraints
  arr J,tJ,phiHatQ,x,v,dx;
  double precision,precisionv;

  //variables that describe the dynamics
  arr Minv,F;

  //variables to store beliefs
  n *= 2;
  arr tmp1,tmp2,tmp3,tmp4,
    a(T,n),invA(T,n,n),
    z(T,n),invZ(T,n,n),
    b(T,n),B(T,n,n),invB(T,n,n),
    r(T,n),R(T,n,n),
    hatq(T,n),hatqq;
  a[0]=q_0;
  invA[0].setDiag(1e10);
  b[0]=q_0;
  B[0].setDiag(1e-10);
  invB[0].setDiag(1e10);
  r[0]=0.;
  R[0]=0.;
  z.setZero();
  invZ.setZero();

  MT::timerStart();
  
  uint i,k,t,dt,t0;
  for(k=0;k<iterations;k++){
    if(!(k&1)){ dt=1; t0=1; }else{ dt=(uint)-1; t0=T-1; }
    for(t=t0;t<T && t>0;t+=dt){
      //decide on a point of linearization (of dynamics & kinematics)
      if(k==0){
	hatq[t]()=b[t-1];
      }else{
	if(stepSmooth){
	  hatq[t]()=(1.-stepSmooth)*b[t]+stepSmooth*hatq[t];
	}else{
	  hatq[t]()=b[t];
	}
      }
      soci.setqv(hatq[t]);
      soci.getMinvF(Minv,F);

      //cout <<"hatq\n" <<hatq[t-1] <<'\n' <<Minv <<'\n' <<F <<endl;

      //compute (a,A)
      inverse_SymPosDef(tmp1,invA[t-1] + R[t-1]);
      setBlockVec  (tmp2, tau2,tau, Minv*F);
      blockTimesVec(tmp3, 1,tau,0,1, tmp1 * (invA[t-1]*a[t-1] + r[t-1]));
      a[t]() = tmp2 + tmp3;
      setBlockDiagMat  (tmp2, invW, Q+tau2*Minv*invH*~Minv);
      blockTransMat(tmp3, 1,tau,0,1, tmp1);
      inverse_SymPosDef(invA[t](), tmp2 + tmp3);

      //cout <<"a\n" <<a[t] <<'\n' <<invA[t] <<endl;

      //decide on a point of linearization (of dynamics & kinematics)
      /*if(k>0 && stepSmooth){
	hatq[t]()=(1.-stepSmooth)*b[t]+stepSmooth*hatq[t];
      }else{
        if(k) hatq[t]()=b[t]; else hatq[t]()=a[t];
      }
      soci.setqv(hatq[t]);
      soci.getMinvF(Minv,F);*/

      //cout <<"hatq\n" <<hatq[t-1] <<'\n' <<Minv <<'\n' <<F <<endl;

      //compute (z,Z)
      if(k && t<T-1){
	inverse_SymPosDef(tmp1,invZ[t+1] + R[t+1]);
	setBlockVec  (tmp2, 0,-tau, Minv*F);
	blockTimesVec(tmp3, 1,-tau,0,1, tmp1 * (invZ[t+1]*z[t+1] + r[t+1]));
	z[t]() = tmp2 + tmp3;
	setBlockDiagMat(tmp2, invW, Q+tau2*Minv*invH*~Minv);
	blockTransMat  (tmp3, 1,-tau,0,1, tmp1);
	inverse_SymPosDef(invZ[t](), tmp2 + tmp3);
      }
      if(k && t==T-1){
        z[t] = b[t];
        //invZ[t].setDiag(1e10); //fixes the end posture!, setZero otherwise
	invZ[t].setDiag(1e-2);
      }

      //cout <<"z\n" <<z[t] <<'\n' <<invZ[t] <<endl;

      //compute (r,R)
      r[t].setZero();
      R[t].setZero();
      for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
        soci.getPhi(phiHatQ,i);
	soci.getJtJ(J,tJ,i);
	soci.getTarget (x,i,t);
	soci.getTargetV(v,i,t);
	soci.getPrecision(precision,i,t);
	soci.getPrecisionV(precisionv,i,t);
	hatqq=hatq[t];  hatqq.resizeCopy(n/2);
        dx = x - phiHatQ + J*hatqq;
	setBlockVec(tmp1, precision*tJ*dx, precisionv*tJ*v);
	setBlockMat(tmp2, precision,0,0,precisionv, tJ*J);
        r[t]() += tmp1;
        R[t]() += tmp2;
      }
      //tmp1=hatq[t];
      //for(i=n/2;i<n;i++) tmp1(i)=0.;
      //r[t]() += R[t] * tmp1;

      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;

      //compute (b,B);
      invB[t] = invA[t] + invZ[t] + R[t];
      inverse_SymPosDef(B[t](), invB[t]);
      b[t] = B[t] * (invA[t]*a[t] + invZ[t]*z[t] + r[t]);

      //cout <<"b\n" <<b[t] <<endl <<B[t] <<endl;

      //soci.displayBelief(b[t],k,t);

      if(k>0 && repeatThreshold){
        double off=sqrDistance(W,b[t],hatq[t]);
	if(off>repeatThreshold){
	  //cout <<t <<" REPEAT: off = " <<off <<endl;
	  t-=dt;
	}
      }
    }

    //display or evaluate
    if(soci.os){
      *soci.os <<std::setw(3) <<k <<" " <<MT::timerRead(false);
      soci.computeTotalCost(b);
    }
    if(soci.gl){
      soci.displayTrajectory(b,display,STRING("bayesian optimization -- iteration "<<k));
    }
  }

  q = b;
}

//! obsolete
void soc::BayesianDynamicFilter_obsolete(SocAbstraction& soci,
					 arr& q,uint iterations,
					 double stepSmooth,double repeatThreshold,
					 uint display){
  uint n=soci.qDim();
  uint T=soci.nTime();
  uint m=soci.nTasks();
  double tau=soci.getTau(), tau2=tau*tau;
  
  //general variables
  arr q0,H,invH,Q;
  soci.getH(H);
  soci.getQ(Q);
  inverse_SymPosDef(invH,H);
  soci.getq0(q0);
  
  //variables that describe the constraints
  arr J,tJ,phiHatQ,x,dx;
  double precision;
  
  //variables that describe the dynamics
  arr Minv,F;
  
  //variables to store beliefs
  arr tmp1,tmp2,tmp3,tmp4,
    a(T,n),invA(T,n,n),
    z(T,n),invZ(T,n,n),
    b(T,n),B(T,n,n),invB(T,n,n),
    r(T,n),R(T,n,n),
    hatq(T,n),hatqd(T,n);
  a[0]=q0;
  invA[0].setDiag(1e10);
  b[0]=q0;
  B[0].setDiag(1e-10);
  invB[0].setDiag(1e10);
  r[0]=0.;
  R[0]=0.;
  z.setZero();
  invZ.setZero();
  hatq[0]=q0;
  hatqd[0]=0.;

  MT::timerStart();
  
  uint i,k,t,dt,t0;
  for(k=0;k<iterations;k++){
    if(!(k&1)){ dt=1; t0=1; }else{ dt=(uint)-1; t0=T-1; }
    for(t=t0;t<T && t>0;t+=dt){
      //decide on a point of linearization (of dynamics & kinematics)
      if(k==0){
	hatq[t]()=b[t-1];
	hatqd[t]() = (hatq[t  ]-hatq[t-1])/tau;
      }else{
	if(stepSmooth){
	  hatq[t]()=(1.-stepSmooth)*b[t]+stepSmooth*hatq[t];
	}else{
	  hatq[t]()=b[t];
	}
	getVelocity(hatqd[t](),hatq,t,tau);
      }
      soci.setqv(hatq[t],hatqd[t]);
      soci.getMinvF(Minv,F);
      
      //cout <<"hatq\n" <<hatq[t] <<"hatqd\n" <<hatqd[t] <<'\n' <<Minv <<'\n' <<F <<endl;
      
      //compute (a,A)
      if(t>1){
        inverse_SymPosDef(tmp1,invA[t-1] + R[t-1]);
        inverse_SymPosDef(tmp2,invA[t-2] + R[t-2]);
	a[t] = tau2*Minv*F;
	a[t]() += 2.*tmp1 *(invA[t-1]*a[t-1] + r[t-1]);
	a[t]() -=    tmp2 *(invA[t-2]*a[t-2] + r[t-2]);
	
	tmp3 = tmp1; //4.*
	//tmp3 += tmp2;
	tmp3 += Q;
	//tmp3 += tau2*Minv*invH*~Minv;
        inverse_SymPosDef(invA[t](), tmp3);
      }else{
        inverse_SymPosDef(tmp1,invA[t-1] + R[t-1]);
	a[t] = tmp1 *(invA[t-1]*a[t-1] + r[t-1]);
        inverse_SymPosDef(invA[t](), tmp1 + Q);
      }
      
      //cout <<"a\n" <<a[t] <<'\n' <<invA[t] <<endl;
      //cout <<"precision a=" <<trace(invA[t]);

      //compute (z,Z)
      if(k && t<T-2){
	inverse_SymPosDef(tmp1,invZ[t+1] + R[t+1]);
	inverse_SymPosDef(tmp2,invZ[t+2] + R[t+2]);
	z[t] = tau2*Minv*F;
	z[t]() += 2.*tmp1 *(invZ[t+1]*z[t+1] + r[t+1]);
	z[t]() -=    tmp2 *(invZ[t+2]*z[t+2] + r[t+2]);
	
	tmp3 = tmp1; //4.*
	tmp3 += tmp2;
	tmp3 += Q;
	tmp3 += tau2*Minv*invH*~Minv;
        inverse_SymPosDef(invZ[t](), tmp3);
      }
      if(k && t==T-2){
        inverse_SymPosDef(tmp1,invZ[t+1] + R[t+1]);
	z[t] = tmp1 *(invZ[t+1]*z[t+1] + r[t+1]);
        inverse_SymPosDef(invZ[t](), tmp1 + Q);
      }
      if(k && t==T-1){
	z[t] = b[t];
	//invZ[t].setDiag(1e10); //fixes the end posture!, use 1e-5 otherwise
	invZ[t].setDiag(1e-2);
      }

      //cout <<"z\n" <<z[t] <<'\n' <<invZ[t] <<endl;
      //cout <<" z=" <<trace(invZ[t]);
      
      //compute (r,R)
      r[t].setZero();
      R[t].setZero();
      for(i=0;i<m;i++) if(soci.isConditioned(i,t)){
	soci.getPhi(phiHatQ,i);
	soci.getJtJ(J,tJ,i);
	soci.getTarget(x,i,t);
	soci.getPrecision(precision,i,t);
	dx = x - phiHatQ; //the J*hatq term is added 5 lines below
	r[t]() += precision * tJ * dx;
	R[t]() += precision * tJ * J;
      }
      r[t]() += R[t] * hatq[t];
      
      //cout <<"r\n" <<r[t] <<endl <<R[t] <<endl;
      //cout <<" r=" <<trace(R[t]);
      
      //compute (b,B);
      invB[t] = invA[t] + invZ[t] + R[t];
      inverse_SymPosDef(B[t](), invB[t]);
      b[t] = B[t] * (invA[t]*a[t] + invZ[t]*z[t] + r[t]);
      
      //cout <<"b\n" <<b[t] <<endl <<B[t] <<endl;
      //cout <<" b=" <<trace(invB[t]) <<endl;
      
      //soci.displayBelief(b[t],k,t);
      
      /*if(k>0 && repeatThreshold){
      double off=sqrDistance(W,b[t],hatq[t]);
      if(off>repeatThreshold){
      //cout <<t <<" REPEAT: off = " <<off <<endl;
      t-=dt;
      }
    }*/
    }
    
    //display or evaluate
    if(soci.os){
      *soci.os <<std::setw(3) <<k <<" " <<MT::timerRead(false);
      soci.computeTotalCost(b);
    }
    if(soci.gl){
      soci.displayTrajectory(b,display,STRING("bayesian optimization -- iteration "<<k));
    }

    //evaluate trajectory
    //double cost1,cost2;
    //getFilterCostMeassure(soci,b,cost1,cost2,soci.os);
  }
  q = b;
}


//===========================================================================
//
// helpers - fwd implementations
//

void setBlockVec(arr& X, double a0,double a1, const arr& x){
  CHECK(x.nd==1,"");
  uint i,n=x.N;
  X.resize(2*n);
  X.setZero();
  if(a0!=0){
    memmove(X.p,x.p,n*x.sizeT);
    if(a0!=1) for(i=0;i<n;i++) X(i) *= a0;
  }
  if(a1!=0){
    memmove(X.p+n,x.p,n*x.sizeT);
    if(a1!=1) for(i=0;i<n;i++) X(n+i) *= a1;
  }
}

void setBlockVec(arr& X,const arr& x1,const arr& x2){
  CHECK(x1.nd==1 && x2.nd==1 && x1.N==x2.N,"");
  uint n=x1.N;
  X.resize(2*n);
  memmove(X.p  ,x1.p,n*x1.sizeT);
  memmove(X.p+n,x2.p,n*x2.sizeT);
}

void setBlockDiagMat(arr& X,const arr& A,const arr& B){
  CHECK(A.nd==2 && B.nd==2 && A.d1==B.d1 && A.d0==B.d0,"");
  uint i; //,n=A.d0;
  X.resize(A.d0+B.d0,A.d1+B.d1);
  X.setZero();
  for(i=0;i<A.d0;i++) memmove(&X(i     ,0   ),&A(i,0),A.d1*A.sizeT);
  for(i=0;i<B.d0;i++) memmove(&X(i+A.d0,A.d1),&B(i,0),B.d1*A.sizeT);
}

void setBlockMat(arr& X, double a00,double a01,double a10,double a11, const arr& x){
  CHECK(x.nd==2 && x.d0==x.d1,"");
  uint i,j,n=x.d0;
  X.resize(2*n,2*n);
  X.setZero();
  if(a00!=0){
    for(i=0;i<n;i++) memmove(&X(i  ,0),&x(i,0),n*x.sizeT);
    if(a00!=1) for(i=0;i<n;i++) for(j=0;j<n;j++) X(  i,  j) *= a00;
  }
  if(a01!=0){
    for(i=0;i<n;i++) memmove(&X(i  ,n),&x(i,0),n*x.sizeT);
    if(a01!=1) for(i=0;i<n;i++) for(j=0;j<n;j++) X(  i,n+j) *= a01;
  }
  if(a10!=0){
    for(i=0;i<n;i++) memmove(&X(n+i,0),&x(i,0),n*x.sizeT);
    if(a10!=1) for(i=0;i<n;i++) for(j=0;j<n;j++) X(n+i,  j) *= a10;
  }
  if(a11!=0){
    for(i=0;i<n;i++) memmove(&X(n+i,n),&x(i,0),n*x.sizeT);
    if(a11!=1) for(i=0;i<n;i++) for(j=0;j<n;j++) X(n+i,n+j) *= a11;
  }
}

void blockTimesVec(arr& Y, double a00,double a01,double a10,double a11, const arr& X){
  CHECK(X.nd==1,"");
  uint i,n=X.N/2;
  CHECK(X.N==2*n,"");
  Y.resize(2*n);
  Y.setZero();
  if(a00!=0) for(i=0;i<n;i++) Y(  i) += a00*X(  i);
  if(a01!=0) for(i=0;i<n;i++) Y(  i) += a01*X(n+i);
  if(a10!=0) for(i=0;i<n;i++) Y(n+i) += a10*X(  i);
  if(a11!=0) for(i=0;i<n;i++) Y(n+i) += a11*X(n+i);
}

void bl_ockTimesMat(arr& Y, double a00,double a01,double a10,double a11, const arr& X){
  CHECK(X.nd==2 && X.d0==X.d1,"");
  uint i,j,n=X.d0/2;
  CHECK(X.d0==2*n,"");
  Y.resize(2*n,2*n);
  Y.setZero();
  if(a00!=0) for(i=0;i<n;i++) for(j=0;j<n;j++){ Y(  i,  j) += a00*X(  i,  j); Y(  i,n+j) += a00*X(  i,n+j); }
  if(a01!=0) for(i=0;i<n;i++) for(j=0;j<n;j++){ Y(  i,  j) += a01*X(n+i,  j); Y(  i,n+j) += a01*X(n+i,n+j); }
  if(a10!=0) for(i=0;i<n;i++) for(j=0;j<n;j++){ Y(n+i,  j) += a10*X(  i,  j); Y(n+i,n+j) += a10*X(  i,n+j); }
  if(a11!=0) for(i=0;i<n;i++) for(j=0;j<n;j++){ Y(n+i,  j) += a11*X(n+i,  j); Y(n+i,n+j) += a11*X(n+i,n+j); }
}

void blockTransMat(arr& Y, double a,double b,double c,double d, const arr& X){
  CHECK(X.nd==2 && X.d0==X.d1,"");
  uint i,j,n=X.d0/2;
  CHECK(X.d0==2*n,"");
  Y.resize(2*n,2*n);
  Y.setZero();
  for(i=0;i<n;i++) for(j=0;j<n;j++)
    Y(  i,  j) += a*a*X(  i,  j) + a*b*(X(  i,n+j)+X(n+i,  j)) + b*b*X(n+i,n+j);
  for(i=0;i<n;i++) for(j=0;j<n;j++)
    Y(n+i,n+j) += d*d*X(n+i,n+j) + c*d*(X(  i,n+j)+X(n+i,  j)) + c*c*X(  i,  j);
  for(i=0;i<n;i++) for(j=0;j<n;j++)
    Y(  i,n+j) += a*c*X(  i,  j) + b*c*X(n+i,  j) + a*d*X(  i,n+j) + b*d*X(n+i,n+j);
  for(i=0;i<n;i++) for(j=0;j<n;j++)
    Y(n+i,  j) += a*c*X(  i,  j) + a*d*X(n+i,  j) + b*c*X(  i,n+j) + b*d*X(n+i,n+j);
}  

double getDynamicCostMeassure(soc::SocAbstraction& soci,arr& q,double& cost1,double& cost2,std::ostream *os){
  //evaluate trajectory
  double cost_t,length=0.;
  cost1=.0;
  cost2=.0;
  double tau=soci.getTau(),tau2=tau*tau;
  arr W,invW,H,invH,Q,Minv,F,tmp1,tmp2,tmp3,tmp4,d;
  soci.getW(W);
  soci.getH(H);
  soci.getQ(Q);
  inverse_SymPosDef(invW,W);
  inverse_SymPosDef(invH,H);
  uint t,T=q.d0;
  for(t=0;t<T;t++){
    soci.setqv(q[t]);
    if(t>0){
      soci.getMinvF(Minv,F);
      blockTimesVec(tmp1, 1,tau,0,1, q[t-1]);
      setBlockVec  (tmp2, tau2,tau, Minv*F);
      setBlockDiagMat(tmp3, invW, Q+tau2*Minv*invH*~Minv);
      inverse_SymPosDef(tmp4,tmp3);
      tmp1=q[t]-(tmp1+tmp2);
      cost2 += scalarProduct(tmp4,tmp1,tmp1);
    }
    if(t>0){ d=q[t]-q[t-1]; d.resizeCopy(q.d1/2); length += ::sqrt(scalarProduct(W,d,d)); }
    cost1 += cost_t = getTaskCost(soci,t);
  }
  if(soci.os){
    *soci.os <<std::setw(3) <<0
        <<"  time " <<MT::timerRead(false)
        <<"  cost1 " <<cost1
        <<"  cost2 " <<cost2
        <<"  length " <<length
        <<"  total-cost " <<cost1+cost2 <<endl;
  }
  return cost1+cost2;
}

double getTransitionLogLikelihood(soc::SocAbstraction& soci,arr& b0,arr& b1){
  double tau=soci.getTau(),tau2=tau*tau;
  arr W,invW,H,invH,Q,Minv,F,tmp1,tmp2,tmp3,tmp4;
  soci.getW(W);
  soci.getH(H);
  soci.getQ(Q);
  soci.getMinvF(Minv,F);
  inverse_SymPosDef(invW,W);
  inverse_SymPosDef(invH,H);
  blockTimesVec(tmp1, 1,tau,0,1, b0);
  setBlockVec  (tmp2, tau2,tau, Minv*F);
  setBlockDiagMat(tmp3, invW, Q+tau2*Minv*invH*~Minv);
  inverse_SymPosDef(tmp4,tmp3);
  tmp1=b1-(tmp1+tmp2);
  return scalarProduct(tmp4,tmp1,tmp1);
  //return logNN(b1,tmp1+tmp2,tmp3);
  return -1.;
}

